# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import os
import django
from django.contrib.auth.models import Group, User
from django.db.models import Q
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
django.setup()
from EPCPyYes.core.v1_2.events import EventType
from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
from rest_framework.test import APITestCase
from django.urls import reverse
from quartet_output import models
from quartet_output.management.commands.create_output_groups import Command

os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.settings'
django.setup()


class ViewTest(APITestCase):
    def setUp(self):
        user = User.objects.create_user(username='testuser',
                                        password='unittest',
                                        email='testuser@seriallab.local')
        Command().handle()
        oag = Group.objects.get(name='Output Access')
        user.groups.add(oag)
        user.save()
        self.client.force_authenticate(user=user)
        self.user = user

    def test_endpoint(self):
        self._create_endpoint()
        url = reverse('end-points-list')
        response = self.client.get(url)
        self.assertIs(response.status_code, 200)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0].get('name'), 'Test EndPoint')

    def test_authentication_info(self):
        self._create_auth()
        url = reverse('authentication-info-list')
        response = self.client.get(url)
        self.assertIs(response.status_code, 200)

    def test_epcis_output_criteria(self):
        self._create_good_ouput_criterion()
        url = reverse('epcis-output-criteria-list')
        response = self.client.get(url)
        self.assertIs(response.status_code, 200)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0].get('action'), 'ADD')

    def test_read_only_epcis_output_criteria(self):
        self._create_good_ouput_criterion()
        url = reverse('read-only-epcis-output-criteria-list')
        response = self.client.get(url)
        self.assertIs(response.status_code, 200)
        self.assertEqual(len(response.data), 1)
        self.assertEqual(response.data[0].get('action'), 'ADD')

    def _create_good_ouput_criterion(self):
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc = models.EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "ADD"
        eoc.event_type = EventType.Transaction.value
        eoc.disposition = Disposition.in_transit.value
        eoc.biz_step = BusinessSteps.shipping.value
        eoc.biz_location = 'urn:epc:id:sgln:305555.123456.0'
        eoc.read_point = 'urn:epc:id:sgln:305555.123456.12'
        eoc.source_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.source_id = 'urn:epc:id:sgln:305555.123456.12'
        eoc.destination_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.destination_id = 'urn:epc:id:sgln:309999.111111.233'
        eoc.authentication_info = auth
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_endpoint(self):
        ep = models.EndPoint()
        ep.urn = 'http://testhost'
        ep.name = 'Test EndPoint'
        ep.save()
        return ep

    def _create_auth(self):
        auth = models.AuthenticationInfo()
        auth.description = 'Unit test auth.'
        auth.username = 'UnitTestUser'
        auth.password = 'UnitTestPassword'
        auth.save()
        return auth

    def _get_test_data(self):
        '''
        Loads the XML file and passes its data back as a string.
        '''
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path) as data_file:
            return data_file.read()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import logging
from typing import List
from EPCPyYes.core.v1_2 import events
from quartet_output.models import EPCISOutputCriteria

logger = logging.getLogger(__name__)

SourceList = List[events.Source]
DestinationList = List[events.Destination]


class EventEvaluation:

    def evaluate_event(self, event: events.EPCISBusinessEvent,
                       epc_output_criteria: EPCISOutputCriteria
                       ):
        '''
        Returns a true or false if the event meets the output criteria.
        :param event: An EPCPyYes event to evaluate.
        :param epc_output_criteria: The criteria to evaluate the event against.
        :return: boolean - True or false.
        '''
        # check the event type
        event_type = self.check_event_type(
            event,
            epc_output_criteria.event_type
        )
        action = self.check_action(
            event, epc_output_criteria.action
        )
        biz_location = self.check_biz_location(
            event,
            epc_output_criteria.biz_location
        )
        disposition = self._check_string(
            event.disposition,
            epc_output_criteria.disposition
        )
        biz_step = self._check_string(
            event.biz_step,
            epc_output_criteria.biz_step
        )
        read_point = self._check_string(
            event.read_point,
            epc_output_criteria.read_point
        )
        source = self.check_sources(
            event,
            epc_output_criteria
        )
        destination = self.check_destinations(
            event,
            epc_output_criteria
        )
        return (event_type and action and biz_location and disposition
                and biz_step and read_point and source and destination)

    def check_sources(self, event: events.EPCISBusinessEvent,
                      epc_output_criteria: EPCISOutputCriteria):
        '''
        Loops through the sources of the event looking for any expected
        source type/id combinations.
        :param event: The event
        :param epc_output_criteria: The output criteria containing the
        expected source definition (id and type)
        :return: True or False - True if there is a matching source type/id
        combination.
        '''
        ret = False
        if epc_output_criteria.source_type and len(event.source_list) > 0:
            for source in event.source_list:
                ret = self._check_string(source.type,
                                         epc_output_criteria.source_type) and \
                      self._check_string(source.source,
                                         epc_output_criteria.source_id)
                if ret:
                    break
        else:
            ret = True
        return ret

    def check_destinations(self, event: events.EPCISBusinessEvent,
                           epc_output_criteria: EPCISOutputCriteria):
        '''
        Loops through the destinations of the event looking for any expected
        destination type/id combinations.
        :param event: The event
        :param epc_output_criteria: The output criteria containing the
        expected destination definition (id and type)
        :return: True or False - True if there is a matching destination type/id
        combination.
        '''
        ret = False
        if epc_output_criteria.destination_type and \
            len(event.destination_list) > 0:
            for destination in event.destination_list:
                ret = self._check_string(
                    destination.type,
                    epc_output_criteria.destination_type
                ) and \
                      self._check_string(
                          destination.destination,
                          epc_output_criteria.destination_id
                      )
                if ret:
                    break
        else:
            ret = True
        return ret

    def check_event_type(self, event: events.EPCISEvent, event_type: str):
        '''
        Checks the event against the expected event type.
        :param event: The event to check
        :param event_type: The expected type
        :return: True or false
        '''
        ret = False if event_type else True
        if event_type:
            if event_type == events.EventType.Transaction.value:
                ret = isinstance(event, events.TransactionEvent)
            elif event_type == events.EventType.Object.value:
                ret = isinstance(event, events.ObjectEvent)
            elif event_type == events.EventType.Aggregation.value:
                ret = isinstance(event, events.AggregationEvent)
            elif event_type == events.EventType.Transformation.value:
                ret = isinstance(event, events.TransformationEvent)
        return ret

    def check_biz_location(self, event, biz_location: str):
        '''
        Checks the event biz_location against the expected one.
        :param event: The event to check
        :param biz_location: The expected biz_location
        :return: True or False
        '''
        ret = False if biz_location else True
        if biz_location:
            ret = (biz_location == event.biz_location)
        return ret

    def check_action(self, event, action: str):
        '''
        Evaluates the event action vs. the expected action
        :param event: The event to evaluate
        :param action: The expected action
        :return: True or False (bool)
        '''
        ret = False if action else True
        if action:
            try:
                ret = (event.action == action)
            except AttributeError:
                # transformation events will not have an action
                # there is only one attribute call in that code
                # so this handler is safe
                pass
        return ret

    def _check_string(self, event_val, criteria_val):
        '''
        Checks an event value against the expected value. If the criterial
        val was not specified will return True since it's not for
        consideration.
        :param event_val: The event value.
        :param criteria_val: The expected value.
        :return: True or False
        '''
        ret = False if criteria_val else True
        if criteria_val:
            ret = (event_val == criteria_val)
        return ret



# -*- coding: utf-8 -*-
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from EPCPyYes.core.v1_2.events import EventType
from django.db import models
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

ACTION_CHOICES = (
    ('ADD', 'Add'),
    ('OBSERVE', 'Observe'),
    ('DELETE', 'Delete')
)

EVENT_CHOICES = (
    (EventType.Object.value, EventType.Object.value),
    (EventType.Transaction.value, EventType.Transaction.value),
    (EventType.Transformation.value, EventType.Transformation.value),
    (EventType.Aggregation.value, EventType.Aggregation.value)
)


class EPCISOutputCriteria(models.Model):
    """
    Based on the contents of an inbound event or groups of events,
    this model can be used to examine the contents of the events and
    determine what rules to execute for messaging.
    """
    name = models.CharField(
        max_length=150,
        verbose_name=_("Name"),
        help_text=_("The name of the endpoint"),
        null=False
    )
    event_type = models.CharField(
        max_length=20,
        verbose_name=_("Event Type"),
        help_text=_("The type of EPCIS event."),
        null=True,
        choices=EVENT_CHOICES
    )
    action = models.CharField(
        max_length=20,
        verbose_name=_("Action"),
        help_text=_("The EPCIS event's ACTION type."),
        null=True,
        choices=ACTION_CHOICES
    )
    biz_step = models.CharField(
        max_length=150,
        verbose_name=_("Business Step (BizStep)"),
        help_text=_("The business step URN.  Can be a CBV value or any custom"
                    "uri.  If CBV it must be exactly as specified in v1.2"),
        null=True
    )
    disposition = models.CharField(
        max_length=150,
        verbose_name=_("Disposition"),
        help_text=_("A Disposition URN- can be CBV or custom."),
        null=True
    )
    read_point = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The read point URN.  Typically representing a sub-site.'),
        verbose_name=_('Read Point')
    )
    biz_location = models.CharField(
        max_length=150,
        null=True,
        help_text=_('The business location URN.  Typically representing '
                    'a site'),
        verbose_name=_('Business Location')
    )
    source_type = models.CharField(
        max_length=150,
        verbose_name=_("Source Type"),
        help_text=_("The type of the source- a CBV 1.2 URI or custom URI."),
        null=True
    )
    source_id = models.CharField(
        max_length=200,
        verbose_name=_("Source ID"),
        help_text=_("A URI that identifies the source specified in the "
                    "Source Type field."),
        null=True
    )
    destination_type = models.CharField(
        max_length=150,
        verbose_name=_("Destination Type"),
        help_text=_(
            "The type of the Destination- a CBV 1.2 URI or custom URI."),
        null=True
    )
    destination_id = models.CharField(
        max_length=200,
        verbose_name=_("Destination ID"),
        help_text=_("A URI that identifies the Destination specified in the "
                    "Destination Type field."),
        null=True
    )
    authentication_info = models.ForeignKey(
        'quartet_output.AuthenticationInfo',
        null=True,
        on_delete=models.SET_NULL,
        verbose_name=_("Authentication Info"),
        help_text=_("The Authentication Info to use."),
    )
    end_point = models.ForeignKey(
        'quartet_output.EndPoint',
        null=False,
        on_delete=models.PROTECT,
        verbose_name=_("End Point"),
        help_text=_("A prtocol-specific endpoint defining where any output"
                    " data will be sent."),
    )

    def clean(self):
        """
        Require that if both source and destination fields are configured if
        one or the other type/id is.
        """
        if self.source_type or self.source_id:
            if not (self.source_type and self.source_id):
                raise ValidationError(_("If either the Source Type or ID "
                                        "are specified then both must be."))
        if self.destination_type or self.destination_id:
            if not (self.destination_type and self.destination_id):
                raise ValidationError(_("If either the Destination Type or ID "
                                        "are specified then both must be."))

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = _('EPCIS Output Criteria')
        verbose_name_plural = _('EPCIS Output Criteria')


class EndPoint(models.Model):
    """
    Defines a generic endpoint.
    """
    name = models.CharField(
        max_length=150,
        verbose_name=_("Name"),
        help_text=_("The name of the endpoint"),
        null=False
    )
    urn = models.CharField(
        max_length=1000,
        verbose_name=_("URL"),
        help_text=_("A uniform resource locator that includes "
                    "protocol information."),
        null=False
    )

    def __str__(self):
        return self.name

class AuthenticationInfo(models.Model):
    """
    Holds data relative to basic auth needed by EndPoints for HTTP and other
    username/password protocols.  Encryption of the password is handled
    through the API layer.  Do not bypass the API when creating users or
    you run the risk of exposing authentication data.
    """
    username = models.CharField(
        max_length=150,
        verbose_name=_("Username"),
        help_text=_("The username."),
        null=False
    )
    password = models.CharField(
        max_length=25,
        verbose_name=_("Password"),
        help_text=_("The password for the user."),
        null=False
    )
    type = models.CharField(
        max_length=25,
        verbose_name=_("Type"),
        help_text=_("The type of authentication to use.  For example, Basic,"
                    " Digest, SSH, etc.  This is optional depending on whether"
                    " or not any steps require this information.")
    )
    description = models.CharField(
        max_length=200,
        verbose_name=_("Description"),
        help_text=_("An optional description."),
        null=True
    )
    private_key = models.TextField(
        verbose_name=_("Private Key"),
        help_text=_("Any private key value if applicable."),
        null=True
    )
    public_key = models.TextField(
        verbose_name=_("Public Key"),
        help_text=_("Any public key info if applicable."),
        null=True
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from quartet_output import models
from rest_framework.serializers import ModelSerializer


class EndPointSerializer(ModelSerializer):
    """
    Default serializer for the EndPoint model.
    """

    class Meta:
        model = models.EndPoint
        fields = '__all__'


class AuthenticationInfoSerializer(ModelSerializer):
    """
    Default serializer for the AuthenticationInfo model.
    """

    class Meta:
        model = models.AuthenticationInfo
        fields = '__all__'
        extra_kwargs = {'password': {'write_only': True}}


class ReadOnlyEPCISOutputCriteriaSerializer(ModelSerializer):
    """
    Default read_only serializer for the EPCISOutputCriteria model.
    """
    end_point = EndPointSerializer(many=False, read_only=True)
    authentication_info = AuthenticationInfoSerializer(many=False,
                                                       read_only=True)

    class Meta:
        model = models.EPCISOutputCriteria
        fields = '__all__'


class EPCISOutputCriteriaSerializer(ModelSerializer):
    """
    Default serializer for the EPCISOutputCriteria model.
    """

    class Meta:
        model = models.EPCISOutputCriteria
        fields = '__all__'



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
import io
import requests
from urllib.parse import urlparse
from enum import Enum
from django.utils.translation import gettext as _
from django.core.files.base import File
from EPCPyYes.core.v1_2 import events
from EPCPyYes.core.v1_2 import template_events
from quartet_capture.rules import RuleContext
from quartet_output import errors
from quartet_output.transport.http import HttpTransportMixin
from quartet_output.models import EPCISOutputCriteria, EndPoint
from quartet_output.parsing import SimpleOutputParser, BusinessOutputParser
from quartet_capture import models, rules, errors as capture_errors
from quartet_capture.tasks import create_and_queue_task
from quartet_epcis.parsing.steps import EPCISParsingStep
from quartet_epcis.db_api.queries import EPCISDBProxy, EntryList
from quartet_epcis.models.choices import EventTypeChoicesEnum


class ContextKeys(Enum):
    """
    Containes Rule Context keys that the steps in this module utilize.

    FILTERED_EVENTS_KEY
    ------------------
    This is used by the `OutputParsingStep` to place any events that match
    the outbound determination parameters set by it's configured
    *EPCIS Output Criteria* step parameter value.  Subsequent Steps in this
    module will look for events under this key to verify if they should
    process the events.  The `OutputParsingStep` will place any matching events
    in this key as `EPCPyYes.core.v1_2.template_event` instances.

    EPCIS_OUTPUT_CRITERIA_KEY
    -------------------------
    When the OutputParsingStep loads, it will put an instance of the
    configured `EPCISOutputCriteria` model that was loaded as a result of
    inspecting the `EPCIS Output Criteria` step parameter for that step.
    This model instance is utilized (typically) by downstream transport
    steps that require a reference to the endpoint, protocol and authentication
    information that is stored in the output criteria.

    AGGREGATION_EVENTS_KEY
    ---------------------
    This is utilized by the `UnpackHierarchy` step.  Any
    EPCPyYes.core.v1_2.template_event.AggregationEvents that were created
    as a result of the events in the `FILTERED_EVENTS_KEY` will be placed under
    this key.  Subsequent Steps will use this key to find the aggregation
    events for further processing, sending, storing, rendering, etc.

    AGGREGATION_EVENTS_KEY
    ---------------------
    This is utilized by the `EPCPyYesParsingStep` step.  Any
    EPCPyYes.core.v1_2.template_event.ObjectEvents that were created
    as a result of the events in the `FILTERED_EVENTS_KEY` will be placed under
    this key.  Subsequent Steps will use this key to find the aggregation
    events for further processing, sending, storing, rendering, etc.

    CREATED_TASK_NAME_KEY
    ---------------------
    When the `CreateOutputTaskStep` creates a new task for deferred processing
    it will store the value here for other interested steps to obtain if
    necessary.
    """
    FILTERED_EVENTS_KEY = 'FILTERED_EVENTS'
    EPCIS_OUTPUT_CRITERIA_KEY = 'EPCIS_OUTPUT_CRITERIA'
    AGGREGATION_EVENTS_KEY = 'AGGREGATION_EVENTS'
    OBJECT_EVENTS_KEY = 'OBJECT_EVENTS'
    OUTBOUND_EPCIS_MESSAGE_KEY = 'OUTBOUND_EPCIS_MESSAGE'
    CREATED_TASK_NAME_KEY = 'CREATED_TASK_NAME'


class OutputParsingStep(EPCISParsingStep):
    """
    A Step that can invokes the `quartet_output.EPCISOutputCriteria`
    parser.  This parser looks for matches between the Step's configured
    *EPCIS Output Criteria* model and any inbound EPCIS events to determine
    if subsequent outbound rule processing is necessary.

    The `OutputParsingStep` will place any events that match up against
    the configured *EPCIS Output Criteria* into the Rule Context's context
    dictionary under the key 'FILTERED_EVENTS'.  This will be a list of
    EPCPyYes template events.

    In addition, since the output criteria contains transport information and
    authentication information, it will place this on the context
    under the EPCIS_OUTPUT_CRITERIA_KEY for any
    downstream steps that need to access the criteria values to send any
    data that was created as a result of the filter step.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self.declared_parameters['EPCIS Output Criteria'] = (
            'The name value of an EPCIS Output Criteria configuration.'
        )
        self.epc_output_criteria = self.get_output_criteria()

    def get_output_criteria(self):
        self.info('Retrieving the Step\'s EPCIS Output Criteria '
                  'parameter value...')
        output_criteria = self.get_parameter(_('EPCIS Output Criteria'),
                                             raise_exception=True)
        self.info(_('EPCIS Output Critieria is set to %s' % output_criteria))
        try:
            return EPCISOutputCriteria.objects.get(
                name=output_criteria
            )
        except EPCISOutputCriteria.DoesNotExist:
            exc = EPCISOutputCriteria.DoesNotExist(
                _('EPCISOutputCriteria with name %s could not be found in the '
                  'database.') % output_criteria
            )
            raise exc

    def execute(self, data, rule_context: rules.RuleContext):
        """
        Calls the SimpleOutputParser which looks for any matches
        between the output criteria and any inbound events.
        :param data: The data to parse.
        :param rule_context: Any context supplied by the Rule.
        :return: None
        """
        # before we start, make sure we make the output criteria available
        # to any downstream steps that need it in order to send data.
        rule_context.context[
            ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value
        ] = self.epc_output_criteria
        # get the parser to use from the parameter value.
        # the loose_enforcement parameter is from the base class
        # `EPCISParsingStep` and determines which parser to use.
        parser_type = SimpleOutputParser if self.loose_enforcement \
            else BusinessOutputParser
        self.info('Parser Type %s', str(parser_type))
        try:
            if isinstance(data, File):
                parser = parser_type(data,
                                     self.epc_output_criteria)
            else:
                parser = parser_type(io.BytesIO(data),
                                     self.epc_output_criteria)
        except TypeError:
            parser = SimpleOutputParser(io.BytesIO(data.encode()),
                                        self.epc_output_criteria)
        self.info(_('Parsing the document...'))
        parser.parse()
        self.info(_('Parsing complete.  %s matching events were found.') %
                  str(len(parser.filtered_events)))
        rule_context.context[
            ContextKeys.FILTERED_EVENTS_KEY.value] = parser.filtered_events


class FilteredEventStepMixin:
    """
    A mixin with some helper functions to deal with common issues.  To use
    this in your own class do the following in your custom step's `execute`
    function:

    .. code-block:: text

        self.rule_context = rule_context

    This will allow the mixin access to the `RuleContext` instance passed to
    every step by the rule engine.
    """

    def get_filtered_events(self, default=[]):
        """
        Will check to see if any prior rules have filtered any events
        for outbound processing.
        :param default: The value to return if no filtered events are found.
        Default is an empty list.
        :return: Will return the events or an empty list.
        """
        return self.rule_context.context.get(
            ContextKeys.FILTERED_EVENTS_KEY.value, default
        )

    def get_epc_list(self, epcis_event: events.EPCISEvent):
        """
        Based on the type of event will return the proper list since
        an object and aggregation event have different names for their
        lists.
        :param epcis_event:
        :return: A list of EPCS.
        """
        if isinstance(epcis_event, events.ObjectEvent):
            ret = epcis_event.child_epcs
        else:
            ret = epcis_event.epc_list
        return ret

    def get_parent_epc(self, epcis_event: events.EPCISEvent,
                       default: str = None):
        '''
        If the inbound event is a transaction or transformation event,
        will return the parent.  Otherwise, default.
        :param epcis_event:
        :return:
        '''
        if isinstance(epcis_event, events.AggregationEvent) or isinstance(
            epcis_event, events.TransformationEvent
        ):
            return epcis_event.parent_id
        else:
            ret = default
        return ret


class UnpackHierarchyStep(rules.Step, FilteredEventStepMixin):
    """
    Will take the top level items from any events and unpack them into
    aggregation events.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self.db_proxy = EPCISDBProxy()

    def execute(self, data, rule_context: rules.RuleContext):
        """
        Ignores the data and looks in the RuleContext for any events in the
        rule context that have been selected for outbound message processing
        under the key defined in the `FILTERED_EVENTS_KEY`.

        If there are events in the context under that key, this Step will
        then
        :param data: The original inbound message to the Rule.
        :param rule_context: The RuleContext instance passed in by the
        rule engine.
        """
        self.out_events = []
        self.rule_context = rule_context
        self.info('Unpack hierarchy step executing...')
        # look in the context for any events that matched output determination
        # criteria set forth in an `EPCISOutputCriteria` model instance.
        epcis_events = self.get_filtered_events()
        # use the db_proxy to get all the events for this.
        self.info('Found %s events that were filtered.', len(epcis_events))
        epcs = []
        for epcis_event in epcis_events:
            if isinstance(epcis_event, events.AggregationEvent) or \
                isinstance(epcis_event, events.TransformationEvent):
                self.warning(
                    _('Skipping a transformation or aggregation event. '
                      'these events can not trigger an unpack action '
                      'in the UnpackHierarchyStep.  Only Transaction '
                      'and Object events may do so.'))
            else:
                epcs = epcs + self.get_epc_list(epcis_event)
                if epcis_event.parent_id:
                    epcs.append(epcis_event.parent_id)

        # use the db proxy to get the EPCPyYes aggregation event history back
        agg_events = self.db_proxy.get_aggregation_events_by_epcs(epcs)
        # add the found events to the context for any downstream steps
        rule_context.context[
            ContextKeys.AGGREGATION_EVENTS_KEY.value] = agg_events

    def on_failure(self):
        pass

    @property
    def declared_parameters(self):
        return super().declared_parameters()


class AddCommissioningDataStep(rules.Step, FilteredEventStepMixin):
    """
    This step will look at the rule context FILTERED_EVENTS_KEY for any filterd
    EPCIS events.  If any are found, this step will use those events to create
    the series of ObjectEvents that created the items and any item children
    in the filtered events.
    """

    def __init__(self, db_task: models.Task, **kwargs):
        super().__init__(db_task, **kwargs)
        self.db_proxy = EPCISDBProxy()

    def execute(self, data, rule_context: RuleContext):
        '''
        Looks for any filtered events and then creates any object events
        associated with the epcs in the filtered event.
        :param data: The rule data (not used by this step)
        :param rule_context: The rule context.
        '''
        # check for filtered events in the rule context
        epcis_events = rule_context.context.get(
            ContextKeys.FILTERED_EVENTS_KEY.value, []
        )
        # set this to use the mixin
        self.rule_context = rule_context
        self.info('%s filtered events have been found. Processing',
                  len(epcis_events))
        for epcis_event in epcis_events:
            self.process_event(epcis_event, rule_context)
        self.info('Processing complete.')

    def process_event(self, epcis_event: events.EPCISBusinessEvent,
                      rule_context: RuleContext):
        '''
        Handles any filtered EPCPyYes events found in the context, unpacks
        their children and adds them to EPCPyYes object event instances.
        Any EPCPyYes events created by this function are put on the rule
        context under the OBJECT_EVENTS_KEY key.
        :param epcis_event: An EPCPyYes event found on the context that
        was filtered.
        :param rule_context: The rule context to add any created events to.
        :return:
        '''
        epcs = self.get_epc_list(epcis_event)
        parent = self.get_parent_epc(epcis_event)
        if parent: epcs.append(parent)
        # find if there are any top-level entries- this is a much more
        # efficient database query
        tops = self.db_proxy.get_top_entries(epcs)
        # if there were any tops, remove them from the epcs list so we don't
        # double our efforts
        [epcs.remove(top.identifier) for top in tops]
        # now that we have all the tops, get all of the entries that have
        # the defined tops
        all_children = self.db_proxy.get_entries_by_tops(
            tops,
            select_for_update=False
        )
        # now get any entries from any of the other epcs that are parent-level
        # if any
        parents = self.db_proxy.get_parent_entries(
            epcs,
            select_for_update=False
        )
        all_children = all_children | self.handle_parent_entries(parents)
        # now find all the commissioning events for these
        all_events = self.db_proxy.get_events_by_entry_list(
            all_children | parents,
            event_type=EventTypeChoicesEnum.OBJECT.value
        )
        self.info('Adding %s Object events to the rule context.',
                  len(all_events))
        rule_context.context[ContextKeys.OBJECT_EVENTS_KEY.value] = all_events

    def handle_parent_entries(self, parent_entries: EntryList):
        '''
        Recursively walks a hierarchy and pulls out epcs for processing.
        :param parent_entries: Entries that represent EPCs that are parent
        EPCs.
        :return: All of the children found for the given entries.
        '''
        # the epcs in the event automatically are assumed to be part of an
        # object event of commissioning so we have them already, now we
        # need to see if there are any "parents" in the child list
        # of the event and add those children as well.
        all_children = self.db_proxy.get_entries_by_parents(parent_entries)
        for child in all_children:
            if child.is_parent:
                all_children = all_children | self.handle_parent_entries(
                    [child])
        return all_children

    def on_failure(self):
        pass

    @property
    def declared_parameters(self):
        return super().declared_parameters()


class EPCPyYesOutputStep(rules.Step, FilteredEventStepMixin):
    """
    Will look for any EPCPyYes events in the context and render them to
    XML or JSON depending on the step parameter configuration.
    """

    def execute(self, data, rule_context: RuleContext):
        """
        Pulls the object, agg, transaction and other events out of the context
        for processing.  See the step parameters
        :param data: The original message (not used by this step).
        :param rule_context: The RuleContext containing any filtered events
        and also any EPCPyYes events that were created by prior steps.
        """
        self.rule_context = rule_context
        append_filtered_events = self.get_boolean_parameter(
            'Append Filtered Events', True)
        prepend_filtered_events = self.get_boolean_parameter(
            'Prepend Filtered Events', False
        )
        oevents = rule_context.context.get(ContextKeys.OBJECT_EVENTS_KEY.value,
                                           [])
        aggevents = rule_context.context.get(
            ContextKeys.AGGREGATION_EVENTS_KEY.value, [])
        if append_filtered_events:
            if prepend_filtered_events:
                all_events = self.get_filtered_events() + oevents + aggevents
            else:
                all_events = oevents + aggevents + self.get_filtered_events()
        else:
            all_events = oevents + aggevents
        if len(all_events) > 0:
            epcis_document = template_events.EPCISEventListDocument(all_events)
            if self.get_boolean_parameter('JSON', False):
                data = epcis_document.render_json()
            else:
                data = epcis_document.render()
            rule_context.context[
                ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value
            ] = data

    def declared_parameters(self):
        return {
            "Append Filtered Events": _('Whether or not to append any events '
                                        'found in the rule context under the '
                                        'FILTERED_EVENTS_KEY to the new '
                                        'message.'),
            "Prepend Filtered Events": _('Whether or not to add the filtered '
                                         'events to the beginning of the '
                                         'message or the end. If this is True,'
                                         ' then all filtered events will be'
                                         'added to the beginning of the EPCIS '
                                         'EventList- otherwise they will be '
                                         'added to the end.'),
            "JSON": _('If set to True then the output message for the EPCPyYEs'
                      'events will be JSON.'),
        }

    def on_failure(self):
        super().on_failure()


class CreateOutputTaskStep(rules.Step):
    '''
    Will look for any data in the rule context under the
    OUTBOUND_EPCIS_MESSAGE_KEY context key.  If data is found there,
    this step will create a new task for the rule engine to process with
    the data found in the context as the data to be processed by the rule.
    To configure what Rule this step will create the new task under,
    set the **Output Rule** step parameter value to the name of a rule to
    execute.
    '''

    def execute(self, data, rule_context: RuleContext):
        '''
        Checks the context for any data under the OUTBOUND_EPCIS_MESSAGE_KEY
        and then creates a task in the rule engine using the *Outbound Rule*
        step parameter.
        :param data: The data to process.
        :param rule_context: The rule context.
        '''
        super().execute(data, rule_context)
        # check the context to see if we have any epcis data to send
        self.info(_('Checking the rule context for any data under the '
                    'OUTBOUND_EPCIS_MESSAGE_KEY context key.'))
        data = rule_context.context.get(
            ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value
        )

        if data:
            self.info(_('Data was found.  Checking for the Output Rule '
                        'parameter in the step parameters.'))
            # get the output rule name
            output_rule_name = self.get_parameter(
                'Output Rule',
                raise_exception=True
            )
            self.info(_('Looking for the EPCISOutputCriteria placed on the '
                        'context by the OutputParsingStep.'))
            # get the epcis output critieria to put in the task parameters
            # output steps need access to the auth and endpoint info therein
            epcis_output_criteria = rule_context.get_required_context_variable(
                ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value
            )
            # create a task parameter with the name of the
            # epcis_output_criteria.  the function below will associate with
            # the task after it is created.
            task_param = models.TaskParameter(
                name='EPCIS Output Criteria',
                value=epcis_output_criteria,
                description=_('The name of the EPCIS Output Criteria to '
                              'use during task processing.')
            )
            # send the data and the rule name and task type over
            # to the rule engine to create a task to process the message
            # using the rule specified in this step's 'Output Rule' parameter
            # create it in a waiting state until after parameters are supplied
            run_immediately = self.get_boolean_parameter('run-immediately',
                                                         default=False)
            task = create_and_queue_task(
                data, output_rule_name,
                'Output',
                task_parameters=[task_param],
                run_immediately=run_immediately
            )
            rule_context.context[ContextKeys.CREATED_TASK_NAME_KEY] = task.name
            self.info('Created a new output task %s with rule %s',
                      task.name, output_rule_name)

    def declared_parameters(self):
        return {
            "Output Rule": _('The name of the rule that will process the '
                             'EPCIS output data created by this step.')
        }

    def on_failure(self):
        pass


class TransportStep(rules.Step, HttpTransportMixin):
    '''
    Uses the transport information within the `EPCISOutputCriteria` placed
    on the context under the EPCIS_OUTPUT_CRITERIA_KEY to send any data that
    was placed on the context under the OUTBOUND_EPCIS_MESSAGE_KEY.
    '''

    def execute(self, data, rule_context: RuleContext):
        # get the task parameters that we rely on
        try:
            self.info(_('Looking for the task parameter with the EPCIS '
                        'Output Name.'))
            param = models.TaskParameter.objects.get(
                task__name=rule_context.task_name,
                name='EPCIS Output Criteria'
            )
            # now see if we can get the output critieria based on the param
            # value
            self.info(_('Found the output param, now looking up the '
                        'EPCIS Output Criteria instance with name %s.'),
                      param.value
                      )
            output_criteria = EPCISOutputCriteria.objects.get(
                name=param.value
            )
            self.info(_('Found output criteria with name %s.'),
                      output_criteria)
            # check the url/urn to see if we support the protocol
            protocol = self._supports_protocol(output_criteria.end_point)
            self.info('Protocol supported.  Sending message to %s.' %
                      output_criteria.end_point.urn)
            self._send_message(data, protocol, rule_context, output_criteria)

        except models.TaskParameter.DoesNotExist:
            raise capture_errors.ExpectedTaskParameterError(
                _('The task parameter with name EPCIS Output Criteria '
                  'could not be found.  This task parameter is required by '
                  'the TransportStep to function correctly.')
            )

    def _send_message(
        self,
        data: str,
        protocol: str,
        rule_context: RuleContext,
        output_criteria: EPCISOutputCriteria
    ):
        '''
        Sends a message using the protocol specified.
        :param protocol: The scheme of the urn in the output_criteria endpoint.
        :param rule_context: The RuleContext contains the data in the
        OUTBOUND_EPCIS_MESSAGE_KEY value from the `ContextKey` class.
        :param output_criteria: The originating output criteria.
        :return: None.
        '''
        content_type = self.get_parameter('content-type', 'application/xml')
        file_extension = self.get_parameter('file-extension', 'xml')
        put_data = self.get_boolean_parameter('put-data')
        if protocol.lower() in ['http', 'https']:
            if not put_data:
                func = self.post_data
            else:
                func = self.put_data
            func(
                data,
                rule_context,
                output_criteria,
                content_type,
                file_extension
            )

    def _supports_protocol(self, endpoint: EndPoint):
        '''
        Inspects the output settings and determines if this step can support
        the protocol or not. Override this to support another or more
        protocols.
        :param EndPoint: the endpoint to inspect
        :return: Returns the supported scheme if the protocol is supported or
        None.
        '''
        parse_result = urlparse(
            endpoint.urn
        )
        if parse_result.scheme.lower() in ['http', 'https']:
            return parse_result.scheme
        else:
            raise errors.ProtocolNotSupportedError(_(
                'The protocol specified in urn %s is not supported by this '
                'step or module.'
            ), endpoint.urn)

    def on_failure(self):
        super().on_failure()

    @property
    def declared_parameters(self):
        return {
            'content-type': 'The content-type to add to the header during any '
                            'http posts, puts, etc. Default is application/'
                            'xml',
            'file-extension': 'The file extension to specify when posting and '
                              'putting data via http. Default is xml'
        }



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from EPCPyYes.core.v1_2 import events as yes_events
from quartet_output.evaluation import EventEvaluation
from quartet_output.models import EPCISOutputCriteria
from quartet_epcis.parsing.parser import QuartetParser
from quartet_epcis.parsing.business_parser import BusinessEPCISParser


class SimpleOutputParser(QuartetParser):
    """
    Inherits from the `BusinessEPCISParser` which, unlike the `QuartetParser`,
    enforces strict business rules within the confines of parsing EPCIS data.
    For example, items that have been decommissioned and are included
    within a message with throw an exception, aggregation is tracked and
    hierarchies are maintained, etc.  For more on the difference
    between these two fundamental parsers see the `quartet_epcis`
    documentation.
    """

    def __init__(
        self,
        stream,
        epcis_output_criteria: EPCISOutputCriteria,
        event_cache_size: int = 1024,
        recursive_decommission: bool = True
    ):
        super().__init__(stream, event_cache_size, recursive_decommission)
        self.epcis_output_criteria = epcis_output_criteria
        self.event_evaluation = EventEvaluation()
        self.filtered_events = []

    def handle_aggregation_event(
        self,
        epcis_event: yes_events.AggregationEvent
    ):
        """
        Handles any aggregation events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_aggregation_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_transaction_event(
        self,
        epcis_event: yes_events.TransactionEvent
    ):
        """
        Handles any transaction events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_transaction_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_object_event(self, epcis_event: yes_events.ObjectEvent):
        """
        Handles any object events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_object_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_transformation_event(
        self,
        epcis_event: yes_events.TransformationEvent
    ):
        """
        Handles any transformation events as they are found, inserts them into
        the database and then inspects them for output determination.
        If an event is determined to be meeting an outbound criteria
        configuration then it is added to this class's `filtered_events`
        list.
        :param epcis_event: The event to insert and inspect.
        """
        super().handle_transformation_event(epcis_event)
        self.evaluate(epcis_event)

    def evaluate(self, epcis_event):
        if self.event_evaluation.evaluate_event(
            epcis_event,
            self.epcis_output_criteria
        ):
            self.filtered_events.append(epcis_event)



class BusinessOutputParser(BusinessEPCISParser):
    """
    Inherits from the `BusinessEPCISParser` which, unlike the `QuartetParser`,
    enforces strict business rules within the confines of parsing EPCIS data.
    For example, items that have been decommissioned and are included
    within a message with throw an exception, aggregation is tracked and
    hierarchies are maintained, etc.  For more on the difference
    between these two fundamental parsers see the `quartet_epcis`
    documentation.
    """

    def __init__(
        self,
        stream,
        epcis_output_criteria: EPCISOutputCriteria,
        event_cache_size: int = 1024,
        recursive_decommission: bool = True
    ):
        super().__init__(stream, event_cache_size, recursive_decommission)
        self.epcis_output_criteria = epcis_output_criteria
        self.event_evaluation = EventEvaluation()
        self.filtered_events = []

    def handle_aggregation_event(
        self,
        epcis_event: yes_events.AggregationEvent
    ):
        """
        Handles any aggregation events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_aggregation_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_transaction_event(
        self,
        epcis_event: yes_events.TransactionEvent
    ):
        """
        Handles any transaction events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_transaction_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_object_event(self, epcis_event: yes_events.ObjectEvent):
        """
        Handles any object events as they are found, inserts them into
        the database and then inspects them for output determination.
        :param epcis_event: The event to insert and inspect.
        :return: None
        """
        super().handle_object_event(epcis_event)
        self.evaluate(epcis_event)

    def handle_transformation_event(
        self,
        epcis_event: yes_events.TransformationEvent
    ):
        """
        Handles any transformation events as they are found, inserts them into
        the database and then inspects them for output determination.
        If an event is determined to be meeting an outbound criteria
        configuration then it is added to this class's `filtered_events`
        list.
        :param epcis_event: The event to insert and inspect.
        """
        super().handle_transformation_event(epcis_event)
        self.evaluate(epcis_event)

    def evaluate(self, epcis_event):
        if self.event_evaluation.evaluate_event(
            epcis_event,
            self.epcis_output_criteria
        ):
            self.filtered_events.append(epcis_event)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.utils.translation import gettext as _
from django.core.management.base import BaseCommand, CommandError
from quartet_capture import models
from quartet_output.models import EndPoint, EPCISOutputCriteria, \
    AuthenticationInfo


class Command(BaseCommand):
    help = _(
        'Loads the quartet_output demonstration and transport rules into '
        'the database.'
    )

    def handle(self, *args, **options):
        endpoint = EndPoint.objects.create(
            name=_('Local Echo Server'),
            urn=_('http://localhost')
        )
        auth = AuthenticationInfo.objects.create(
            username=_('Test User'),
            password=_('Test Password'),
            type='Digest',
            description=_('A test user for the example rule.')
        )
        output = EPCISOutputCriteria.objects.create(
            name=_('Test Transaction Criteria'),
            action='ADD',
            event_type='Transaction',
            biz_location='urn:epc:id:sgln:305555.123456.0',
            end_point=endpoint,
            authentication_info=auth
        )
        rule = models.Rule.objects.create(
            name=_('EPCIS Output Filter'),
            description=_('Will inspect inbound messages for output '
                          'processing.')
        )
        parse_step = models.Step.objects.create(
            name=_('Inspect EPCIS'),
            description=_(
                'Parse and insepect EPCIS events using output criteria.'),
            step_class='quartet_output.steps.OutputParsingStep',
            order=1,
            rule=rule
        )
        models.StepParameter.objects.create(
            name='EPCIS Output Criteria',
            step=parse_step,
            value='Test Transaction Criteria',
            description=_(
                'This is the name of the EPCIS Output Criteria record to use.')

        )
        models.Step.objects.create(
            name=_('Add Commissioning Data'),
            description=_(
                'Adds commissioning events for filtered EPCs and their children.'),
            step_class='quartet_output.steps.AddCommissioningDataStep',
            order=2,
            rule=rule
        )
        models.Step.objects.create(
            name=_('Add Aggregation Data'),
            description=_(
                'Adds aggregation events for included EPCs in any filtered events.'),
            step_class='quartet_output.steps.UnpackHierarchyStep',
            order=3,
            rule=rule
        )
        models.Step.objects.create(
            name=_('Render EPCIS XML'),
            description=_(
                'Pulls any EPCPyYes objects from the context and creates an XML message'),
            step_class='quartet_output.steps.EPCPyYesOutputStep',
            order=4,
            rule=rule
        )
        output_step = models.Step.objects.create(
            name=_('Queue Outbound Message'),
            description=_('Creates a Task for sending any outbound data'),
            step_class='quartet_output.steps.CreateOutputTaskStep',
            order=5,
            rule=rule
        )
        models.StepParameter.objects.create(
            step = output_step,
            name='Output Rule',
            value='Transport Rule'
        )
        trule = models.Rule.objects.create(
            name=_('Transport Rule'),
            description=_('An output Rule for any data filtered by EPCIS Output Criteria '
                          'rules.')
        )
        sdstep = models.Step.objects.create(
            name=_('Send Data'),
            description=_('This will send the task message using the source EPCIS Output '
                          'Critria EndPoint and Authentication Info.'),
            step_class='quartet_output.steps.TransportStep',
            order=1,
            rule=trule
        )
        models.StepParameter.objects.create(
            step=sdstep,
            name='Output Rule',
            value='Transport Rule',
            description=_('The rule that will handle any tasks created by this '
                        'step.')
        )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

class BaseOutputError(Exception):
    def __init__(self, *args: object, **kwargs: object) -> None:
        if len(args) > 1:
            message = args[0] % tuple(args[1:])
        else:
            message = args[0]
        super().__init__(message, **kwargs)

class ParameterNotFoundError(BaseOutputError):
    '''
    Should be raised when an expected parameter is not found.
    '''
    pass

class RuleNotFound(BaseOutputError):
    '''
    Thrown if an outbound rule can not be located to process an outbound
    message.
    '''
    pass

class ExpectedContextVariableError(BaseOutputError):
    '''
    Thrown if the Rule Context does not contain an expected context variable.
    '''
    pass

class ProtocolNotSupportedError(BaseOutputError):
    '''
    Thrown if the Rule Context does not contain an expected context variable.
    '''
    pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.

from rest_framework.routers import DefaultRouter
from quartet_output import viewsets

router = DefaultRouter()

router.register(
    r'end-points',
    viewsets.EndPointViewSet,
    base_name='end-points'
)
router.register(
    r'authentication-info',
    viewsets.AuthenticationInfoViewSet,
    base_name='authentication-info'
)
router.register(
    r'epcis-output-criteria',
    viewsets.EPCISOutputCriteriaViewSet,
    base_name='epcis-output-criteria'
)
router.register(
    r'read-only-epcis-output-criteria',
    viewsets.ReadOnlyCriteriaViewSet,
    base_name='read-only-epcis-output-criteria'
)

urlpatterns = router.urls



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from quartet_capture import models
from io import StringIO
from logging import getLogger
import requests
from requests.auth import HTTPBasicAuth, HTTPDigestAuth, HTTPProxyAuth
from quartet_output.models import EndPoint, EPCISOutputCriteria
from quartet_capture.rules import Step, RuleContext
from quartet_output import __version__
logger = getLogger(__name__)

user_agent = 'quartet-output/{0}'.format(
    __version__
)

class HttpTransportMixin:
    '''
    Add to steps that need to support sending http messages.
    '''

    def post_data(self, data: str, rule_context: RuleContext,
                  output_criteria: EPCISOutputCriteria,
                  content_type='application/xml',
                  file_extension='xml',
                  http_put=False):
        '''
        :param data_context_key: The key within the rule_context that contains
         the data to post.  If being invoked from the internals of this
         module this is usually the OUTBOUND_EPCIS_MESSSAGE_KEY value of the
         `quartet_output.steps.ContextKeys` Enum.
        :param output_criteria: The output criteria containing the connection
        info.
        :return: The response.
        '''
        data_stream = data
        file_name = '{0}.{1}'.format(rule_context.task_name, file_extension)
        logger.debug('Posting data with urn %s and file_name %s and '
                     'extension %s.',
                     output_criteria.end_point.urn,
                     file_name,
                     file_extension)
        if not http_put:
            func = requests.post
            files = {'file': data_stream}
        else:
            func = requests.put
            file_name = '{0}.{1}'.format(rule_context.task_name,
                                         file_extension)
            files = {'file': (file_name, data_stream)}
        response = func(
            output_criteria.end_point.urn,
            files,
            auth=self.get_auth(output_criteria),
            headers={'content-type': content_type, 'user-agent': user_agent}
        )
        return response

    def put_data(self, data: str, rule_context: RuleContext,
                 output_criteria: EPCISOutputCriteria,
                 content_type='application/xml',
                 file_exension='xml'):
        '''
        :param data: The data to PUT.
        :param output_criteria: The output criteria containing the connection
        info.
        :return: The response.
        '''
        return self.post_data(data, rule_context, output_criteria,
                              content_type, file_exension, http_put=True)

    def get_auth(self, output_criteria):
        """
        Get's the authentication method and credentials from the
        EPCISOutputCriteria record.
        :param output_criteria: An EPCISOutputCriteria model instance.
        :return: A `requests.auth.HTTPBasicAuth` or `HTTPProxyAuth`
        """
        auth_info = output_criteria.authentication_info
        auth = None
        if auth_info:
            auth_type = auth_info.type or ''
            if 'digest' in auth_type.lower():
                auth = HTTPBasicAuth
            elif 'proxy' in auth_type.lower():
                auth = HTTPProxyAuth
            else:
                auth = HTTPBasicAuth
            auth = auth(auth_info.username, auth_info.password)
        return auth



# -*- coding: utf-8
from __future__ import unicode_literals, absolute_import

import django

DEBUG = True
USE_TZ = True

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "n&kjfd5rc6wh-=y-&6472a-zn*w2%uat8jwwd99jq4%52hc^-h"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

ROOT_URLCONF = "tests.urls"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "quartet_output",
    "quartet_capture",
    "quartet_epcis"
]

SITE_ID = 1

if django.VERSION >= (1, 10):
    MIDDLEWARE = ()
else:
    MIDDLEWARE_CLASSES = ()

MEDIA_ROOT='/tmp'

REST_FRAMEWORK = {
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.TokenAuthentication',
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication'
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
        'rest_framework.permissions.DjangoModelPermissions'
    ),
}

##### un-comment to enable local logging to standard output ######
# import logging
# import logging.config
# logging.basicConfig(
#     level=logging.DEBUG,
#     handlers=[logging.StreamHandler()],
# )
# LOGGING = {
#     'version': 1,
#     'disable_existing_loggers': False,
#     'handlers': {
#         'file': {
#             'level': 'DEBUG',
#             'class': 'logging.StreamHandler',
#         },
#     },
#     'loggers': {
#         'django': {
#             'handlers': ['file'],
#             'level': 'INFO',
#             'propagate': False,
#         },
#     },
# }

try:
    from tests.local_settings import *
except ImportError:
    pass



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2018 SerialLab Corp.  All rights reserved.
from django.utils.translation import gettext as _
from django.core.management.base import BaseCommand
from django.contrib.auth.models import Group, Permission
from django.db.models import Q


class Command(BaseCommand):
    help = _(
        'Creates the default QU4RTET output group.'
    )

    def handle(self, *args, **options):
        group, created = Group.objects.get_or_create(
            name='Output Access'
        )
        if created:
            permissions = Permission.objects.filter(
                Q(codename__endswith='_epsisoutputcriteria') and
                Q(codename__endswith='_authenticationinfo') and
                Q(codename__endswith='_endpoint')
            )
            group.permissions.set(
                permissions
            )
            group.save()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from django.conf.urls import url, include

from quartet_output.urls import urlpatterns as quartet_output_urls

urlpatterns = [
    url(r'^', include(quartet_output_urls)),
]




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_quartet_output
------------

Tests for `quartet_output` models module.
"""
import os
from django.core.exceptions import ValidationError
from EPCPyYes.core.v1_2.events import EventType
from EPCPyYes.core.v1_2.CBV.dispositions import Disposition
from EPCPyYes.core.v1_2.CBV.business_steps import BusinessSteps
from quartet_epcis.parsing.business_parser import BusinessEPCISParser
from quartet_capture.models import Rule, Step, StepParameter, Task
from quartet_capture.tasks import execute_rule, execute_queued_task
from quartet_output.steps import SimpleOutputParser, ContextKeys
from quartet_output.models import EPCISOutputCriteria
from django.test import TestCase

from quartet_output import models


class TestQuartetOutput(TestCase):

    def test_bad_source(self):
        oc = self._create_good_ouput_criterion()
        oc.source_id = None
        with self.assertRaises(ValidationError):
            oc.clean()

    def test_bad_destination(self):
        oc = self._create_good_ouput_criterion()
        oc.destination_id = None
        with self.assertRaises(ValidationError):
            oc.clean()

    def test_rule_with_pallet_agg(self):
        '''
        Should process the EPCIS but not filter out a full pallet hierarchy.
        :return:
        '''
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_three_events.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)

    def test_rule_with_agg_comm(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            self.assertIsNotNone(
                context.context.get(
                    ContextKeys.EPCIS_OUTPUT_CRITERIA_KEY.value)
            )

    def test_rule_with_agg_comm_output(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        db_rule2= self._create_transport_rule()
        self._create_transport_step(db_rule2)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_comm_output_put(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        self._create_task_step(db_rule)
        db_rule2= self._create_transport_rule()
        self._create_transport_step(db_rule2, put_data=True)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_one_event.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            task_name = context.context[ContextKeys.CREATED_TASK_NAME_KEY]
            execute_queued_task(task_name=task_name)
            task = Task.objects.get(name=task_name)
            self.assertEqual(task.status, 'FINISHED')

    def test_rule_with_agg_mulit_comm(self):
        self._create_good_ouput_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        self._create_comm_step(db_rule)
        self._create_epcpyyes_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        # prepopulate the db
        self._parse_test_data('data/commission_three_events.xml')
        self._parse_test_data('data/nested_pack.xml')
        data_path = os.path.join(curpath, 'data/ship_pallet.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                3,
                "There should be three filtered events."
            )
            for event in context.context[
                ContextKeys.AGGREGATION_EVENTS_KEY.value]:
                if event.parent_id in ['urn:epc:id:sgtin:305555.3555555.1',
                                       'urn:epc:id:sgtin:305555.3555555.2']:
                    self.assertEqual(len(event.child_epcs), 5)
                else:
                    self.assertEqual(len(event.child_epcs), 2)
            self.assertIn(ContextKeys.OUTBOUND_EPCIS_MESSAGE_KEY.value,
                          context.context,
                          "No EPCIS message was found in the rule context.")
            self.assertIn(ContextKeys.OBJECT_EVENTS_KEY.value,
                          context.context)
            self.assertEqual(
                len(context.context[ContextKeys.OBJECT_EVENTS_KEY.value]), 3,
                "There should be three object events."
            )
            for oevent in context.context[
                ContextKeys.OBJECT_EVENTS_KEY.value]:
                self.assertIn(len(oevent.epc_list), [10, 2, 1],
                              "One of the object events in the context is "
                              "malformed.")

    def test_rule_no_agg(self):
        '''
        Should process the EPCIS but not filter out any aggregation events.
        :return:
        '''
        # create output criteria that does not result in the
        # matching of any events.
        self._create_bad_criterion()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        self._create_output_steps(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            context = execute_rule(data_file.read().encode(), db_task)
            self.assertEqual(
                len(context.context[ContextKeys.AGGREGATION_EVENTS_KEY.value]),
                0,
                "There should be no filtered events."
            )

    def test_full_bad_rule(self):
        oec = self._create_good_ouput_criterion()
        oec.name = "Bad Name"
        oec.save()
        db_rule = self._create_rule()
        self._create_step(db_rule)
        db_task = self._create_task(db_rule)
        curpath = os.path.dirname(__file__)
        data_path = os.path.join(curpath, 'data/epcis.xml')
        with open(data_path, 'r') as data_file:
            try:
                execute_rule(data_file.read().encode(), db_task)
            except EPCISOutputCriteria.DoesNotExist:
                pass

    def _create_good_ouput_criterion(self):
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "ADD"
        eoc.event_type = EventType.Transaction.value
        eoc.disposition = Disposition.in_transit.value
        eoc.biz_step = BusinessSteps.shipping.value
        eoc.biz_location = 'urn:epc:id:sgln:305555.123456.0'
        eoc.read_point = 'urn:epc:id:sgln:305555.123456.12'
        eoc.source_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.source_id = 'urn:epc:id:sgln:305555.123456.12'
        eoc.destination_type = 'urn:epcglobal:cbv:sdt:location'
        eoc.destination_id = 'urn:epc:id:sgln:309999.111111.233'
        eoc.authentication_info = auth
        eoc.end_point = endpoint
        eoc.save()
        return eoc

    def _create_endpoint(self):
        ep = models.EndPoint()
        ep.urn = 'http://testhost'
        ep.name = 'Test EndPoint'
        ep.save()
        return ep

    def _create_auth(self):
        auth = models.AuthenticationInfo()
        auth.description = 'Unit test auth.'
        auth.username = 'UnitTestUser'
        auth.password = 'UnitTestPassword'
        auth.save()
        return auth

    def _create_bad_criterion(self):
        eoc = EPCISOutputCriteria()
        eoc.name = "Test Criteria"
        eoc.action = "DELETE"
        eoc.event_type = EventType.Transaction.value
        endpoint = self._create_endpoint()
        auth = self._create_auth()
        eoc.end_point = endpoint
        eoc.authentication_info = auth
        eoc.save()
        return eoc

    def _parse_data(self, output_criteria):
        curpath = os.path.dirname(__file__)
        parser = SimpleOutputParser(
            os.path.join(curpath, 'data/epcis.xml'),
            output_criteria
        )
        parser.parse()
        parser.clear_cache()

    def _create_rule(self):
        rule = Rule()
        rule.name = 'output-test'
        rule.description = 'output test rule'
        rule.save()
        return rule

    def _create_transport_rule(self):
        rule = Rule()
        rule.name = 'Transport Rule'
        rule.description = 'Attempts to send data using transport step(s).'
        rule.save()
        return rule

    def _create_transport_step(self, rule, put_data='False'):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Transport'
        step.step_class = 'quartet_output.steps.TransportStep'
        step.description = 'Sends test data.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'run-immediately'
        step_parameter.value = 'True'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'put-data'
        step_parameter.value = put_data
        step_parameter.save()

    def _create_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 1
        step.name = 'Output Determination'
        step.step_class = 'quartet_output.steps.OutputParsingStep'
        step.description = 'unit test step'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'EPCIS Output Criteria'
        step_parameter.value = 'Test Criteria'
        step_parameter.save()

    def _create_output_steps(self, rule):
        step = Step()
        step.rule = rule
        step.order = 2
        step.name = 'UnpackHierarchies'
        step.step_class = 'quartet_output.steps.UnpackHierarchyStep'
        step.description = 'unit test unpacking step'
        step.save()

    def _create_comm_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 3
        step.name = 'CreateCommissioning'
        step.step_class = 'quartet_output.steps.AddCommissioningDataStep'
        step.description = 'unit test commissioning step'
        step.save()

    def _create_epcpyyes_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 4
        step.name = 'Create EPCIS'
        step.step_class = 'quartet_output.steps.EPCPyYesOutputStep'
        step.description = 'Creates EPCIS XML or JSON and inserts into rule' \
                           'context.'
        step.save()

    def _create_task(self, rule):
        task = Task()
        task.rule = rule
        task.name = 'unit test task'
        task.save()
        return task

    def _create_task_step(self, rule):
        step = Step()
        step.rule = rule
        step.order = 5
        step.name = 'Create Output Task'
        step.step_class = 'quartet_output.steps.CreateOutputTaskStep'
        step.description = 'Looks for any EPCIS data on the context and ' \
                           'then, if found, creates a new output task using ' \
                           'the configured Output Rule step parameter.'
        step.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'Output Rule'
        step_parameter.value = 'Transport Rule'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()
        step_parameter = StepParameter()
        step_parameter.step = step
        step_parameter.name = 'run-immediately'
        step_parameter.value = 'True'
        step_parameter.description = 'The name of the rule to create a new ' \
                                     'task with.'
        step_parameter.save()

    def _parse_test_data(self, test_file='data/epcis.xml',
                         parser_type=BusinessEPCISParser,
                         recursive_decommission=False):
        curpath = os.path.dirname(__file__)
        if isinstance(parser_type, BusinessEPCISParser):
            parser = parser_type(
                os.path.join(curpath, test_file),
                recursive_decommission=recursive_decommission
            )
        else:
            parser = parser_type(
                os.path.join(curpath, test_file),
            )
        message_id = parser.parse()
        print(parser.event_cache)
        parser.clear_cache()
        return message_id, parser

    def tearDown(self):
        pass
